import os
import time
import psycopg2
import boto3
import json
import logging
from psc_orgkey import decode_orgkey

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Log environment variable retrieval
aws_region = os.environ.get("AWS_REGION")
s3_bucket_name = os.environ.get("S3_BUCKET_NAME")

rds_hostname = os.environ.get("RDS_HOSTNAME")
rds_password = os.environ.get("RDS_PASSWORD")
rds_username = os.environ.get("RDS_USERNAME")
rds_port = os.environ.get("RDS_PORT")
rds_db_name = os.environ.get("RDS_DB_NAME")


# rds_hostname = "localhost"
# rds_password = "postgres"
# rds_username = "postgres"
# rds_port = 5444
# rds_db_name = "policyservice"


logger.info("Retrieved environment variables:")
logger.info(f"AWS_REGION: {aws_region}")
logger.info(f"RDS_HOSTNAME: {rds_hostname}")
logger.info("RDS_PASSWORD: ********")  # Mask password
logger.info(f"RDS_USERNAME: {rds_username}")
logger.info(f"RDS_PORT: {rds_port}")
logger.info(f"RDS_DB_NAME: {rds_db_name}")
logger.info(f"S3_BUCKET_NAME: {s3_bucket_name}")

folder_name = f"policy-exclusions-export/{time.strftime('%Y-%m-%d_%H-%M-%S')}"


def export_exclusions_data():
    """
    Export policy exclusions data to s3 bucket
    """
    try:
        logger.info("Connecting to the RDS database...")
        conn = psycopg2.connect(
            host=rds_hostname,
            port=rds_port,
            database=rds_db_name,
            user=rds_username,
            password=rds_password,
        )
        cursor = conn.cursor()
        logger.info("Connected to the RDS database.")

        policies = get_policies(cursor)
        logger.info(f"Exporting {len(policies)} policies...")

        export_data = {
            "policies": policies
        }

        write_to_s3(export_data)
    except Exception as e:
        logger.error(
            "Error connecting to the RDS database or storing data in S3: %s", e
        )
    finally:
        # Close the cursor and connection
        if cursor is not None:
            cursor.close()

        if conn is not None:
            conn.close()


def get_rule_criteria_dict(cursor):
    """
    Get criteria id to rule criteria dict
    """
    rule_criteria_attributes = get_rule_criteria_attributes(cursor)

    logger.info("Getting criteria id to rule criteria...")

    cursor.execute("SELECT * FROM policyservice.rule_criteria;")
    rule_criteria_dict = dict()

    for rule_criteria in cursor:
        criteria_id = rule_criteria[0]
        rule_criteria = {
            "type": rule_criteria[1],
            "attributes": rule_criteria_attributes[criteria_id],
        }

        rule_criteria_dict[criteria_id] = rule_criteria

    return rule_criteria_dict


def get_rule_criteria_attributes(cursor):
    """
    Get rule criteria id to attributes dict
    """
    logger.info("Getting rule criteria id to attributes...")

    cursor.execute("SELECT * FROM policyservice.rule_criteria_expression;")
    rule_criteria_attributes = dict()

    for rule_criteria_expression in cursor:
        rule_criteria = {
            "process_name": rule_criteria_expression[1],
            "values": json.loads(rule_criteria_expression[2]),
        }

        key = rule_criteria_expression[3]

        if key in rule_criteria_attributes:
            rule_criteria_attributes[key].append(rule_criteria)
        else:
            rule_criteria_attributes[key] = [rule_criteria]

    return rule_criteria_attributes


def get_bundle_action_id_to_rule_exclusions_dict(cursor):
    """
    Get bundle action id to rule exclusions dict
    """
    logger.info("Getting bundle action id to rule exclusions...")

    rule_exclusion_criteria_relations_dict = get_rule_exclusion_criteria_relations_dict(
        cursor
    )
    rule_criteria_dict = get_rule_criteria_dict(cursor)

    cursor.execute(
        "SELECT id, bundle_action_id, platform, type, apply_to_descendent_processes, comments FROM policyservice.rc_exception;"
    )
    rule_exclusions_dict = {}

    for rule_exclusion_result in cursor:
        rule_exclusion = {}

        if rule_exclusion_result[-2] is not None:
            rule_exclusion["type"] = rule_exclusion_result[3]

        if rule_exclusion_result[-1] is not None:
            rule_exclusion["apply_to_descendent_processes"] = rule_exclusion_result[4]

        if rule_exclusion_result[5] is not None:
            rule_exclusion["comments"] = rule_exclusion_result[5]

        if rule_exclusion_result[0] in rule_exclusion_criteria_relations_dict:
            rule_exclusion["criteria"] = []
            for criteria_id in rule_exclusion_criteria_relations_dict[
                rule_exclusion_result[0]
            ]:
                rule_exclusion["criteria"].append(rule_criteria_dict[criteria_id])

        """
        Add rule exclusion to dict if not exists by bundle action, else append to existing array
        """
        if (
            rule_exclusion_result[1] in rule_exclusions_dict
            and rule_exclusion_result[2]
            in rule_exclusions_dict[rule_exclusion_result[1]]
        ):
            rule_exclusions_dict[rule_exclusion_result[1]][
                rule_exclusion_result[2]
            ].append(rule_exclusion)
        else:
            if rule_exclusion_result[1] not in rule_exclusions_dict:
                rule_exclusions_dict[rule_exclusion_result[1]] = {}
            rule_exclusions_dict[rule_exclusion_result[1]][rule_exclusion_result[2]] = [
                rule_exclusion
            ]

    return rule_exclusions_dict


def get_rule_exclusion_criteria_relations_dict(cursor):
    """
    get rule exception to rule criteria relation dict
    """
    logger.info("Getting rule exception to rule criteria relation...")

    cursor.execute("SELECT * FROM policyservice.rc_exception_criteria_relation;")
    rule_exclusion_criteria_relations_dict = {}

    """
    Add rule criteria relation to dict if not exists by rule criteria, else append to existing array
    """
    for rule_exclusion_criteria_relation in cursor.fetchall():
        if (
            rule_exclusion_criteria_relation[0]
            in rule_exclusion_criteria_relations_dict
        ):
            rule_exclusion_criteria_relations_dict[
                rule_exclusion_criteria_relation[0]
            ].append(rule_exclusion_criteria_relation[1])
        else:
            rule_exclusion_criteria_relations_dict[
                rule_exclusion_criteria_relation[0]
            ] = [rule_exclusion_criteria_relation[1]]

    return rule_exclusion_criteria_relations_dict


def get_policy_id_to_bundle_action_dict(cursor, bundle_action_ids):
    """
    Get org key and policy id to bundle action dict
    """
    logger.info("Getting bundle actions...")

    cursor.execute(
        "SELECT id, bundle_id, SPLIT_PART(scope,':', 3) AS org_key, SPLIT_PART(scope,':', 4) AS policy_id FROM policyservice.rc_bundle_action  WHERE scope LIKE 'psc:policy:%';"
    )
    policy_id_to_bundle_action_dict = {}

    for bundle_action_result in cursor:
        if (bundle_action_result[0] in bundle_action_ids):
            logger.info(f"Getting bundle_action {bundle_action_result[0]}")

            bundle_action = {
                "id": bundle_action_result[0],
                "bundle_id": bundle_action_result[1],
                "org_key": bundle_action_result[2],
                "policy_id": bundle_action_result[3],
            }

            key = f"{bundle_action_result[2]}:{bundle_action_result[3]}"

            if key in policy_id_to_bundle_action_dict:
                policy_id_to_bundle_action_dict[key].append(
                    bundle_action
                )
            else:
                policy_id_to_bundle_action_dict[key] = [bundle_action]

    return policy_id_to_bundle_action_dict


def get_bundles_dict(cursor):
    """
    Get bundle id to bundles dict
    """
    cursor.execute(
        "SELECT id, title, default_scope, category FROM policyservice.rc_bundle"
    )
    bundle_dict = {}

    for bundle_result in cursor:
        logger.info(f"Getting bundle details {bundle_result[0]}")
        bundle = {
            "id": bundle_result[0],
            "title": bundle_result[1],
            "default_scope": bundle_result[2],
            "category": bundle_result[3],
        }

        bundle_dict[bundle_result[0]] = bundle

    return bundle_dict


def get_policies(cursor):
    rc_bundles_dict = get_bundles_dict(cursor)
    bundle_action_id_to_rule_exclusions_dict = get_bundle_action_id_to_rule_exclusions_dict(cursor)
    org_id_policy_id_to_bundle_action_dict = get_policy_id_to_bundle_action_dict(cursor, bundle_action_id_to_rule_exclusions_dict.keys())

    policies = []

    for org_id_policy_id in org_id_policy_id_to_bundle_action_dict.keys():
        logger.info("Getting policy data for %s", org_id_policy_id)

        bundle_actions = org_id_policy_id_to_bundle_action_dict[org_id_policy_id]

        rule_configs = []
        for bundle_action in bundle_actions:
            bundle = rc_bundles_dict[bundle_action["bundle_id"]]
            rule_config = {
                "title": bundle["title"],
                "inherited_from": bundle["default_scope"],
                "category": bundle["category"]
            }

            if bundle_action["id"] in bundle_action_id_to_rule_exclusions_dict:
                rule_config["exclusions"] = bundle_action_id_to_rule_exclusions_dict[bundle_action["id"]]

            rule_configs.append(rule_config)

        decoded_org_key = decode_orgkey(bundle_action["org_key"])

        policy = {
            "id": bundle_action["policy_id"],
            "org_id": decoded_org_key.org_id,
            "environment": decoded_org_key.env,
            "org_key": bundle_action["org_key"],
            "rule_configs": rule_configs,
        }
        policies.append(policy)

    return policies


def write_to_s3(data):
    logger.info("Uploading data.json to s3 bucket %s.", s3_bucket_name)
    s3_client = boto3.client("s3")
    s3_client.put_object(
        Bucket=s3_bucket_name,
        Key=f"{folder_name}/data.json",
        Body=json.dumps(data),
        ContentType="application/json",
    )
    logger.info("Data stored successfully in S3.")


if __name__ == "__main__":
    export_exclusions_data()
